"""MongoDB manager.
"""

import pymongo
from pymongo import MongoClient
from bson.objectid import ObjectId


def objectid_to_str(object_id):
  return str(object_id)


def str_to_objectid(id_str):
  return ObjectId(id_str)


class MongoDBManager(object):
  """MongoDB manager.
  """

  def __init__(self):
    self.client = None
    self.db = None

  def __del__(self):
    self.disconnect()

  def is_connected(self):
    """Check if connection is valid.
    """
    if self.client == None or \
    self.db == None:
      return False
    else:
      return True

  def connect(self,
              user=None,
              pwd=None,
              host=None,
              port=None,
              db_name=None,
              use_replica=False):
    """Connect to mongodb server.

    Args:
      user: user to login.
      pwd: pwd to login.
      host: server host.
      port: port to use.
      db_name: database to use.
      collection_name: collection to use.
    """
    if self.db is None:
      # make connection string.
      if host is None:
        db_url = "mongodb://127.0.0.1:27017"
      else:
        if not use_replica:
          db_url = "mongodb://{}:{}@{}:{}/{}".format(user, pwd, host, port,
                                                     db_name)
        else:
          db_url = "mongodb+srv://{}:{}@{}/{}".format(user, pwd, host, db_name)
      # get client.
      print("connecting to mongodb: {}".format(db_url))
      self.client = MongoClient(db_url)
      if db_name is not None:
        self.db = self.client[db_name]
        print("using db: {}".format(db_name))
      else:
        self.db = self.client["test"]
      print("mongodb connected")

  def delete_db(self, db_name):
    """Delete database.
    """
    self.client.drop_database(db_name)
    print("removed db: {}".format(db_name))

  def delete_collection(self, collection_name):
    """delete collection for current db.
    """
    collection = self.db[collection_name]
    collection.drop()
    print("current collection dropped.")

  def disconnect(self):
    """disconnect from server.
    """
    if self.client is not None:
      self.client.close()
      self.client = None
      self.db = None

  def create_index(self, collection_name, index_attribute):
    """Create index on current collection.

    Args:
      index_attribute: attribute name to index.
    """
    collection = self.db[collection_name]
    collection.create_index(index_attribute)
    print("index {} created.".format(index_attribute))

  def add_items(self, collection_name, items):
    """Add a new document.

    Args:
      items: documents to insert.
    Returns:
      number of documents inserted.
    """
    collection = self.db[collection_name]
    res = collection.insert_many(items)
    return len(res.inserted_ids)

  def get_items(self,
                collection_name,
                query,
                limit=0,
                offset=0,
                sort_key=None,
                sort_order=1):
    """Retrieve items given conditions.

    Args:
      sort_order: descending if 0, otherwise asecnding.
    """
    collection = self.db[collection_name]
    sort_order = pymongo.ASCENDING if sort_order == 1 else pymongo.DESCENDING
    items = list(
        collection.find(
            query, limit=limit, offset=offset, sort=[(sort_key, sort_order)]))
    return items

  def count_items(self, collection_name, query):
    """Count item number.
    """
    collection = self.db[collection_name]
    num_items = collection.count_documents(query)
    return num_items

  def delete_items(self, collection_name, query):
    """Delete items based on query.

    Returns:
      number of deleted documents.
    """
    collection = self.db[collection_name]
    res = collection.delete_many(query)
    return res.deleted_count

  def update_items(self, collection_name, query, update_op):
    """Update items based on query.
    """
    collection = self.db[collection_name]
    collection.update_many(query, update_op)



"""OOP interface for S3.
"""

import io
import os

import boto3
from tinyimage import TinyImage


class S3Manager(object):
  """Class for managing s3 tasks.
  """

  def __init__(self, access_key, secret_key, region="us-east-1"):
    """Create s3 client.
    """
    session = boto3.Session(
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        region_name=region)
    self.client = session.client("s3")

  def create_bucket(self, bucket_name):
    """Create new bucket.
    """
    bucket = self.client.create_bucket(Bucket=bucket_name)
    assert "Location" in bucket, "Bucket creation failed"

  def get_object(self, bucket_name, obj_key):
    """Get object binary data from bucket.

    Args:
      obj_key: object key.
    Returns:
      object data.
    """
    res = self.client.get_object(Bucket=bucket_name, Key=obj_key)
    obj_data = res["Body"].read()
    return obj_data

  def upload_file(self, filepath, bucket_name, file_key):
    """Upload a file to bucket.

    Returns:
      file url.
    """
    self.client.upload_file(filepath, bucket_name, file_key)

  def upload_object(self,
                    obj_binary,
                    bucket_name,
                    obj_key,
                    content_type,
                    enable_download=False,
                    make_public=False):
    """Upload binary object data.

    Args:
      obj_binary: binary encoding of object data.
      bucket_name: name of bucket.
      obj_key: key of object.
      content_type: MIME type of the content, e.g. image/jpeg, text/json.
      enable_download: whether to make the object open or download when accessing the link.

    Returns:
      url to the object.
    """
    self.client.put_object(
        ACL="public-read" if make_public else "private",
        Body=obj_binary,
        Bucket=bucket_name,
        Key=obj_key,
        ContentType=content_type,
        ContentDisposition="attachment" if enable_download else "")
    return self.generate_object_url(
        bucket_name, obj_key, is_public=make_public)

  def upload_img_file(self, filepath, bucket_name, file_key):
    """Upload image file.
    """
    _, img_ext = os.path.splitext(os.path.split(filepath)[1])
    upload_args = {}
    if img_ext in [".jpg", ".jpeg"]:
      upload_args["ContentType"] = "image/jpeg"
    elif img_ext == ".png":
      upload_args["ContentType"] = "image/png"
    else:
      raise Exception("invalid image type")
    self.client.upload_file(
        filepath, bucket_name, file_key, ExtraArgs=upload_args)
    return self.generate_presigned_url(bucket_name, file_key, 3600)

  def upload_img_bin(self,
                     bucket_name,
                     img_bin,
                     img_format="png",
                     key_prefix="",
                     make_public=False):
    """Upload image binary data to bucket.

    Key is named: prefix/img_sha.format.

    Args:
      img_bin: image binary data.
      img_format: type of image, only support jpy or png.
      key_prefix: prefix of image key, used to create folders.
      make_public: whether to make it publically accessible.
    Returns:
      image object key.
    """
    img_format = img_format.lower()
    assert img_format in ["png", "jpg",
                          "jpeg"], "img_format must be either png, jpg, jpeg."
    # make sure no duplicate images are saved.
    img_obj = TinyImage(img_bin=img_bin)
    key_prefix = key_prefix.strip("/")
    img_key_name = "{}/{}.{}".format(key_prefix,
                                     img_obj.get_base64_sha_encoding(),
                                     img_format)
    img_key_name = img_key_name.strip("/")
    upload_args = {}
    if img_format in ["jpg", "jpeg"]:
      upload_args["ContentType"] = "image/jpeg"
    if img_format == "png":
      upload_args["ContentType"] = "image/png"
    if make_public:
      upload_args["ACL"] = "public-read"
    self.client.upload_fileobj(
        io.BytesIO(img_bin), bucket_name, img_key_name, ExtraArgs=upload_args)
    if make_public:
      return img_key_name
    else:
      self.generate_presigned_url()

  def upload_img_base64(self,
                        bucket_name,
                        img_base64,
                        img_format="jpg",
                        key_prefix="",
                        make_public=False):
    """Save image data to bucket.

    sha1 is used as key for image data.

    Args:
      img_base64: base64 string of image.
    Returns:
      image sha hash as key.
    """
    img_obj = TinyImage(img_base64=img_base64)
    img_url = self.upload_img_bin(bucket_name, img_obj.to_binary(), img_format,
                                  key_prefix, make_public)
    return img_url

  def generate_object_url(self,
                          bucket_name,
                          obj_key,
                          is_public=False,
                          expire_secs=604800):
    """Generate url for a given object.

    Args:
      bucket_name: name of bucket.
      obj_key: key of object.
      is_public: if the object is public.
      expire_secs: seconds to expire the url.

    Returns:
      url for the target object.
    """
    if is_public:
      url = "{}/{}/{}".format(self.client.meta.endpoint_url, bucket_name,
                              obj_key)
    else:
      url = self.client.generate_presigned_url(
          ClientMethod="get_object",
          Params={
              "Bucket": bucket_name,
              "Key": obj_key
          },
          ExpiresIn=expire_secs)
    return url



"""Generic interface for web apis.
"""

import abc

from flask import Flask
from flask_cors import CORS

from tornado.wsgi import WSGIContainer
from tornado.httpserver import HTTPServer
from tornado.ioloop import IOLoop


def make_success_res_obj(data, code=200):
  """Create a successful response object.

  Args:
    data: any object to attach to the result.
  """
  return {"status": {"code": code, "msg": "Ok"}, "data": data}


def make_error_res_obj(err_msg, code=500):
  """Create an error response object.

  Args:
    err_msg: error message string.
  """
  return {"error": {"code": code, "msg": err_msg}}


class WebAPIBase(object):
  """Template class for web api services.
  """
  __metaclass__ = abc.ABCMeta

  def __init__(self, api_name):
    self.name = api_name
    self.app = Flask(__name__)
    CORS(self.app, resources={r"/*": {"origins": "*"}})
    self.started = False
    self.add_resource("/", ["GET"], self.index)
    self.add_resource("/info", ["GET"], self.info)

  def add_resource(self, path, method_list, func):
    """Add resources to api.

    Args:
      path: e.g. "/info/".
      method_list: e.g. ["GET", "POST"].
      func: function.
    """
    self.app.add_url_rule(path, methods=method_list, view_func=func)

  def index(self):
    return "Welcome to {} API.".format(self.name)

  def info(self):
    pass

  @abc.abstractmethod
  def prepare(self):
    """Prepare resources before starting the api.
    """
    pass

  def start(self, debug=False, port=5000):
    """Start running the api.
    """
    if not self.started:
      self.prepare()
    if debug:
      self.app.run(debug=True, port=port)
    else:
      http_server = HTTPServer(WSGIContainer(self.app))
      http_server.listen(port)
      print("service on")
      IOLoop.instance().start()


"""Shared services for aws.
"""

import json
import boto3


# TODO(jiefeng): to finish function.
def read_config_fn(config_fn):
  """Read configuration from file.
  """
  with open(config_fn, "r") as f:
    credential = json.load(f)
    return credential


VALID_AWS_SERVICES = ["dynamodb", "s3"]


class AWSService(object):
  """Base class for any aws service.
  """
  session = None
  client = None
  resource = None

  def init_aws(self, service_name, access_key, secret_key, region):
    """Create resource and client.
    """
    assert service_name in VALID_AWS_SERVICES, "invalid aws service name"
    self.session = boto3.Session(
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        region_name=region)
    try:
      self.client = self.session.client(service_name)
    except:
      self.client = None
    try:
      self.resource = self.session.resource(service_name)
    except:
      self.resource = None


"""General tools.
"""

import datetime
import glob
import logging
import os
import random
import uuid
import sys


def append_url_params(base_url, param_dict):
  """Append parameters to a base url.

  Args:
    base_url: the basic url, end with ?
    param_dict: dictionary of parameters.
  Returns:
    a complete url.
  """
  url = base_url
  # indicate if a parameter is attached.
  started = False
  for key, val in param_dict.iteritems():
    if val is not None:
      url = url + "&" if started else url
      url += "{}={}".format(key, val)
      started = True
  return url


def split_train_val_test(labels,
                         train_ratio=0.7,
                         val_ratio=0.0,
                         test_ratio=0.3):
  """split a dataset into train, val and test subsets.

  The split is balanced across classes.

  Args:
    labels: a list of label ids.
    train_ratio: percentage of samples for training.
    val_ratio: percentage of samples for validation.
    test_ratio: percentage of samples for test.
  Returns:
    list of sample ids for train, val, test sets.
  """
  assert (train_ratio + val_ratio +
          test_ratio == 1), "percentages must add up to 1"
  train_ids = []
  val_ids = []
  test_ids = []
  # split for each class evenly.
  unique_labels = set(labels)
  for label in unique_labels:
    indices = [i for i, v in enumerate(labels) if v == label]
    num = len(indices)
    train_num = int(num * train_ratio)
    val_num = int(num * val_ratio)
    random.shuffle(indices)
    train_ids += indices[:train_num]
    val_ids += indices[train_num:train_num + val_num]
    test_ids += indices[train_num + val_num:]
  return train_ids, val_ids, test_ids


def list_img_files(img_root_dir, img_exts=None, has_cate_subfolder=True):
  """Get image file names from a folder.

  Supported image file format: jpg, jpeg, png, bmp.

  Args:
    img_root_dir: root directory of image files.
    img_exts: target image extensions, e.g. [*.png, *.jpg].
    has_cate_subfolder: whether root contains subfolders for categories.
  Returns:
    image file names and labels if category exists.
  """
  img_fns = []
  all_exts = ["*.png", "*.jpg", "*.jpeg", "*.bmp"]
  if img_exts is None:
    img_exts = all_exts
  else:
    for img_ext in img_exts:
      assert img_ext in all_exts, "{} file type not supported".format(img_ext)
  if has_cate_subfolder:
    cate_dirs = os.listdir(img_root_dir)
    cate_dirs = [os.path.join(img_root_dir, x) for x in cate_dirs]
    cate_dirs = [x for x in cate_dirs if os.path.isdir(x)]
    img_labels = []
    label_names = {}
    for cate_id, cur_cate_dir in enumerate(cate_dirs):
      cur_cate_name = os.path.basename(cur_cate_dir.strip("/"))
      label_names[cate_id] = cur_cate_name
      # find all image files.
      all_img_fns = []
      for img_ext in img_exts:
        cur_ext_fns = glob.glob(os.path.join(cur_cate_dir, img_ext))
        all_img_fns += cur_ext_fns
      img_labels += [cate_id] * len(all_img_fns)
      img_fns += all_img_fns
    return img_fns, img_labels, label_names
  else:
    for img_ext in img_exts:
      cur_ext_fns = glob.glob(os.path.join(cur_cate_dir, img_ext))
      img_fns += cur_ext_fns
    return img_fns


def get_logger(name, log_fn=None):
  """Get a logger that outputs to log_fn and stdout.

  Args:
    name: name of the logger.
    log_fn: file to save log data.
  Returns:
    a logger object. use as logger.info, logger.error.
  """
  logger = logging.getLogger(name)
  logger.setLevel(logging.INFO)
  logger.addHandler(logging.StreamHandler())
  if log_fn is None:
    log_fn = "./{}.log".format(name)
  fh = logging.FileHandler(log_fn, mode="w")
  logger.addHandler(fh)
  return logger


def get_datenow_str():
  """Get a properly formatted string for current date and time.
  """
  now = datetime.datetime.now()
  now_str = now.strftime("%Y_%b_%d_%H_%M_%S")
  return now_str


def get_detailed_error_msg(ex_msg, ex_info):
  """Format error message to include details.

  Args:
    ex_msg: exception message.
    ex_info: exception info returned by sys.exc_info()
  Returns:
    a message string including error, file and line number.
  """
  _, exc_obj, exc_tb = ex_info
  fname = exc_tb.tb_frame.f_code.co_filename
  return "error: {}, file: {}, loc: {}".format(ex_msg, fname, exc_tb.tb_lineno)


def get_jsonpath_val(json_obj, path_str):
  """Get json attribute value from a path.

  Args:
    json_obj: json object.
    path_str: attribute path as key1.key2.key3
  Returns:
    value of the given path, if not valid, return None.
  """
  # separate path into keys.
  keys = path_str.split(".")
  cur_val = None
  for cur_key in keys:
    if cur_key in json_obj:
      cur_val = json_obj[cur_key]
    else:
      cur_val = None
      break
  return cur_val


def gen_uuid4():
  """Generate uuid4.
  """
  return str(uuid.uuid4())


def gen_uuid1():
  """Generate uuid1.
  """
  return str(uuid.uuid1())


def list_files(target_dir, fn_exts):
  """List all files match given extensions.

  Match both upper and lower cases.

  Args:
    fn_exts: a list of file extension in the form of "*.jpg".
  
  Returns:
    a list of found files.
  """
  all_exts = []
  for ext in fn_exts:
    all_exts.append(ext.lower())
    all_exts.append(ext.upper())
  all_exts = set(all_exts)
  all_fns = []
  for cur_ext in all_exts:
    cur_fns = glob.glob(os.path.join(target_dir, cur_ext))
    all_fns.extend(cur_fns)
  return all_fns


"""OOP interface for DynamoDB.
"""

import boto3


class DynamoDBManager(object):
  """Class for managing dynamodb tasks.
  """

  def __init__(self, access_key, secret_key, region="us-east-1"):
    """Create dynamodb client.
    """
    session = boto3.Session(
        aws_access_key_id=access_key,
        aws_secret_access_key=secret_key,
        region_name=region)
    self.resource = session.resource("dynamodb")

  def __get_table(self, table_name):
    """Retrieve table object.
    """
    return self.resource.Table(table_name)

  def get_table_item_count(self, table_name):
    """Get how many items are in the current table.
    """
    table = self.resource.Table(table_name)
    assert table.table_status == "ACTIVE", "table is not available"
    return table.item_count

  def create_table(self,
                   table_name,
                   partition_key,
                   sort_key=None,
                   read_cap=1,
                   write_cap=1):
    """Create a table.
    """
    # create one.
    key_schema = [{"AttributeName": partition_key, "KeyType": "HASH"}]
    attributes = [{"AttributeName": partition_key, "AttributeType": "S"}]
    if sort_key is not None:
      key_schema.append({"AttributeName": sort_key, "KeyType": "RANGE"})
      attributes.append({{"AttributeName": sort_key, "AttributeType": "S"}})
    table = self.resource.create_table(
        TableName=table_name,
        KeySchema=key_schema,
        # Only need to define keys type
        AttributeDefinitions=attributes,
        ProvisionedThroughput={
            "ReadCapacityUnits": read_cap,
            "WriteCapacityUnits": write_cap
        })
    print("table {} created.".format(table_name))

  def add_items(self, table_name, item):
    """Write attributes to current table.

    Args:
      item: name value pairs.
    """
    table = self.__get_table(table_name)
    with table.batch_writer() as batch:
      for cur_item in items:
        batch.put_item(Item=cur_item)

  def get_item(self, table_name, key_dict, attributes_to_get=None):
    """Get attributes from table.

    Args:
      key_dict: dictionary of primary key, e.g. {"id": ""}.
      attributes_to_get: projection expression to specify attributes.
    Returns:
      retrieved item with specified attributes.
    """
    table = self.__get_table(table_name)
    if attributes_to_get is None:
      res = table.get_item(Key=key_dict)
    else:
      res = table.get_item(Key=key_dict, AttributesToGet=attributes_to_get)
    if "Item" in res:
      return res["Item"]
    else:
      return None

  def scan_items(self, table_name, limit=10, cont_key=None):
    """Scan table and return items.

    Args:
      cont_key: key used to continue scan.
    Returns:
      current batch of items, continue_key (None if not exist).
    """
    table = self.__get_table(table_name)
    if cont_key is not None:
      response = table.scan(Limit=limit, ExclusiveStartKey=cont_key)
    else:
      response = table.scan(Limit=limit)
    cont_key = None
    if "LastEvaluatedKey" in response:
      cont_key = response["LastEvaluatedKey"]
    return response["Items"], cont_key

  def update_item(self, table_name, key_dict, attr_name, new_attr_val):
    """Update item value.

    Args:
      attr_name: name of attribute.
      new_attr_val: value for the attribut to set.

    Only support SET now.
    """
    table = self.__get_table(table_name)
    table.update_item(
        Key=key_dict,
        UpdateExpression="set {}=:newval".format(attr_name),
        ExpressionAttributeValues={":newval": new_attr_val})



from setuptools import setup, find_packages, find_namespace_packages

# only used for dev.
requirements = []
test_requirements = []
try:
  from pipenv.project import Project
  from pipenv.utils import convert_deps_to_pip
  pfile = Project(chdir=False).parsed_pipfile
  requirements = convert_deps_to_pip(pfile['packages'], r=False)
  requirements = [r for r in requirements if "-e" not in r]
  print(requirements)
  test_requirements = convert_deps_to_pip(pfile['dev-packages'], r=False)
  test_requirements = [r for r in test_requirements if "-e" not in r]
  print(test_requirements)
except:
  pass

setup(
    name="projectowl",
    version="0.3.15",
    description="Utility library for building python app",
    url="https://github.com/flyfj/owl.git",
    author="Jie Feng",
    author_email="jiefeng@perceptance.io",
    package_dir={"": "src"},
    packages=find_namespace_packages(where="src"),
    install_requires=requirements,
    tests_require=test_requirements,
    include_package_data=True,
    zip_safe=False)
